/* -*- mode: c++; coding: utf-8; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4; show-trailing-whitespace: t -*- vim:fenc=utf-8:ft=tcl:et:sw=4:ts=4:sts=4

  This file is part of the Feel library

  Author(s): Christophe Prud'homme <christophe.prudhomme@feelpp.org>
       Date: 2005-11-16

  Copyright (C) 2005,2006 EPFL
  Copyright (C) 2007,2008 Université Joseph Fourier (Grenoble I)

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 3.0 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
/**
   \file importergmsh.hpp
   \author Christophe Prud'homme <christophe.prudhomme@feelpp.org>
   \date 2005-11-16
 */

#ifndef __ImporterGmsh_H
#define __ImporterGmsh_H 1

#include <iostream>
#include <fstream>
#include <iomanip>
#include <algorithm>

#include <feel/feelcore/worldcomm.hpp>
#include <feel/feelfilters/importer.hpp>
#include <feel/feelfilters/gmshenums.hpp>
#include <boost/algorithm/string/trim.hpp>

// Gmsh
#include <GModel.h>
#include <MElement.h>

// from Gmsh
void SwapBytes(char *array, int size, int n);

namespace Feel
{
namespace detail
{
class GMSHPoint
{
public:
    Eigen::Vector3d x;
    Eigen::Vector2d uv;
    int id;
    bool onbdy;
    bool parametric;
    int gdim,gtag;

    EIGEN_MAKE_ALIGNED_OPERATOR_NEW

    GMSHPoint()
        :
        x( Eigen::Vector3d::Zero() ),
        uv( Eigen::Vector2d::Zero() ),
        id( 0 ),
        onbdy( false ),
        parametric( false ),
        gdim( 0 ),
        gtag( 0 )
        {}
    GMSHPoint(GMSHPoint const& p )
        :
        x( p.x ),
        uv( p.uv ),
        id ( p.id ),
        onbdy( p.onbdy ),
        parametric( p.parametric ),
        gdim( p.gdim ),
        gtag( p.gtag )
        {}
};
struct GMSHElement
{
    GMSHElement()
        :
        num( 0 ),
        type( MSH_PNT ),
        physical( 0 ),
        elementary( 0 ),
        numPartitions( 1 ),
        partition( 0 ),
        ghosts(),
        is_on_processor( false ),
        is_ghost( false ),
        ghost_partition_id( -1 ),
        parent( 0 ),
        dom1(0), dom2(0),
        numVertices(0),
        indices()
        {}

    GMSHElement( int n,
                 int t,
                 int p,
                 int e,
                 int _numPartitions,
                 int _partition,
                 std::vector<int> const& _ghosts,
                 int _parent,
                 int _dom1, int _dom2,
                 int _numVertices,
                 std::vector<int> const& _indices )
    :
        num( n ),
        type( t ),
        physical( p ),
        elementary( e ),
        numPartitions( _numPartitions ),
        partition( _partition ),
        ghosts( _ghosts ),
        is_on_processor( false ),
        is_ghost( false ),
        ghost_partition_id( partition ),
        parent( _parent ),
        dom1(_dom1), dom2(_dom2),
        numVertices( _numVertices ),
        indices( _indices )
        {
            int rank = Environment::worldComm().localRank();
            if ( Environment::worldComm().globalSize() == 1 )
            {
                is_on_processor = true;
                is_ghost = false;
            }
            else if ( rank == partition )
            {
                is_on_processor = true;
                is_ghost = false;
            }
            else
            {
                // is the element a ghost cell
                // look into ghosts if 'partition' is present
                auto it = std::find( ghosts.begin(), ghosts.end(), rank );
                if ( it != ghosts.end() )
                {
                    is_on_processor = true;
                    is_ghost = true;
                    ghost_partition_id = partition;
                }
            }
        }

   GMSHElement( GMSHElement const& g )
    :
        num( g.num ),
        type( g.type ),
        physical( g.physical ),
        elementary( g.elementary ),
        numPartitions( g.numPartitions ),
        partition( g.partition ),
        ghosts( g.ghosts ),
        is_on_processor( g.is_on_processor ),
        is_ghost( g.is_ghost ),
        ghost_partition_id( g.ghost_partition_id ),
        parent( g.parent ),
        dom1( g.dom1 ), dom2( g.dom2 ),
        numVertices( g.numVertices ),
        indices( g.indices )
        {}

    bool isOnProcessor() const { return is_on_processor; }
    bool isGhost() const { return is_ghost; }
    int ghostPartitionId() const { return ghost_partition_id; }

    template<typename IteratorBegin,typename IteratorEnd>
    bool isIgnored(IteratorBegin it, IteratorEnd en ) const { return std::find( it, en, physical ) != en; }

    int num;
    int type;
    int physical;
    int elementary;

    //! partitioning info
    int numPartitions;
    int partition;
    std::vector<int> ghosts;
    bool is_on_processor;
    bool is_ghost;
    int ghost_partition_id;

    int parent;
    int dom1, dom2;

    // vertices
    int numVertices;
    std::vector<int> indices;


};
}
/**
 * \class ImporterGmsh
 * \brief gmsh importer class
 *
 * the importer concept follows the visitor pattern
 *
 * \code
 * typename Mesh2D<LinearTetra> mesh_type;
 * mesh_type mesh;
 *
 * ImporterGmsh<mesh_type> import( "mesh.msh");
 * mesh.accept( import );
 * \endcode
 *
 * \ingroup Importer
 * @author Christophe Prud'homme
 */

template<typename MeshType>
class ImporterGmsh
    :

public Importer<MeshType>

{
    typedef Importer<MeshType> super;
public:

    /**
     * setElementRegionAsPhysicalRegion(bool parameter)
     *
     * change reading for importing specific meshes for which the gmsh reader
     * consider the Physical flag as null
     */
    void setElementRegionAsPhysicalRegion( const bool param )
    {
        M_use_elementary_region_as_physical_region=param;
    }

    /** @name Typedefs
     */
    //@{

    typedef typename super::mesh_type mesh_type;
    typedef typename super::point_type point_type;
    typedef typename super::node_type node_type;
    typedef typename super::edge_type edge_type;
    typedef typename super::face_type face_type;
    typedef typename super::element_type element_type;

    typedef typename mesh_type::face_iterator face_iterator;
    //@}

    //@{
    /** @name Constants
     */
    static const uint16_type npoints_per_edge = ( edge_type::numVertices*edge_type::nbPtsPerVertex+
            edge_type::numEdges*edge_type::nbPtsPerEdge+
            edge_type::numFaces*edge_type::nbPtsPerFace );

    static const uint16_type npoints_per_face = ( face_type::numVertices*face_type::nbPtsPerVertex+
            face_type::numEdges*face_type::nbPtsPerEdge+
            face_type::numFaces*face_type::nbPtsPerFace );

    static const uint16_type npoints_per_element = element_type::numPoints;
    //@}

    /** @name Constructors, destructor
     */
    //@{

    ImporterGmsh( WorldComm const& _worldcomm = Environment::worldComm() )
        :
        super( GMSH, _worldcomm ),
        _M_version( FEELPP_GMSH_FORMAT_VERSION ),
        M_use_elementary_region_as_physical_region( false )
    {
        this->setIgnorePhysicalName( "FEELPP_GMSH_PHYSICALNAME_IGNORED" );
        //showMe();
    }

    explicit ImporterGmsh( std::string const& _fname, std::string _version = FEELPP_GMSH_FORMAT_VERSION,
                           WorldComm const& _worldcomm = Environment::worldComm() )
        :
        super( _fname, GMSH, _worldcomm ),
        _M_version( _version ),
        M_use_elementary_region_as_physical_region( false )
    {
        this->setIgnorePhysicalName( "FEELPP_GMSH_PHYSICALNAME_IGNORED" );
        //showMe();
    }
    ImporterGmsh( ImporterGmsh const & i )
        :
        super( i ),
        _M_version( i._M_version ),
        M_use_elementary_region_as_physical_region( false ),
        _M_ignorePhysicalGroup( i._M_ignorePhysicalGroup ),
        _M_ignorePhysicalName( i._M_ignorePhysicalName )
    {
        this->setIgnorePhysicalName( "FEELPP_GMSH_PHYSICALNAME_IGNORED" );
        //showMe();
    }
    ~ImporterGmsh()
    {}

    //@}

    /** @name Operator overloads
     */
    //@{


    //@}

    /** @name Accessors
     */
    //@{

    /**
     * @return the file format version
     */
    std::string version() const
    {
        return _M_version;
    }

    /**
     * \return true if the element is on processor or is a ghost cell
     *         true if a ghost cell
     */
    boost::tuple<bool,boost::tuple<bool,int> > isElementOnProcessor( std::vector<int> const& tags ) const;

    //@}

    /** @name  Mutators
     */
    //@{

    void setVersion( std::string const& version )
    {
        _M_version = version;
    }

    void setIgnorePhysicalGroup( int i )
    {
        _M_ignorePhysicalGroup.insert( i );
    }
    void setIgnorePhysicalName( std::string s )
    {
        _M_ignorePhysicalName.insert( s );
    }

    //@}

    /** @name  Methods
     */
    //@{

    void visit( mesh_type* mesh );

    void showMe() const;

    //@}



protected:

private:

    void addPoint( mesh_type* mesh, Feel::detail::GMSHElement const& __e, int & __idGmshToFeel );
    void addPoint( mesh_type* mesh, Feel::detail::GMSHElement const& __e, int & __idGmshToFeel, mpl::int_<1> );
    void addPoint( mesh_type* /*mesh*/, Feel::detail::GMSHElement const& /*__e*/, int & /*__idGmshToFeel*/, mpl::int_<2> );
    void addPoint( mesh_type* /*mesh*/, Feel::detail::GMSHElement const& /*__e*/, int & /*__idGmshToFeel*/, mpl::int_<3> );

    void addEdge( mesh_type* mesh, Feel::detail::GMSHElement const& __e, int & __idGmshToFeel );
    void addEdge( mesh_type* mesh, Feel::detail::GMSHElement const& __e, int & __idGmshToFeel, mpl::int_<1> );
    void addEdge( mesh_type* mesh, Feel::detail::GMSHElement const& __e, int & __idGmshToFeel, mpl::int_<2> );
    void addEdge( mesh_type* /*mesh*/, Feel::detail::GMSHElement const& /*__e*/, int & /*__idGmshToFeel*/, mpl::int_<3> );

    void addFace( mesh_type* mesh, Feel::detail::GMSHElement const& __e, int & __idGmshToFeel );
    void addFace( mesh_type* /*mesh*/, Feel::detail::GMSHElement const& /*__e*/, int & /*__idGmshToFeel*/, mpl::int_<1> );
    void addFace( mesh_type* mesh, Feel::detail::GMSHElement const& __e, int & __idGmshToFeel, mpl::int_<2> );
    void addFace( mesh_type* mesh, Feel::detail::GMSHElement const& __e, int & __idGmshToFeel, mpl::int_<3> );

    void addVolume( mesh_type* mesh, Feel::detail::GMSHElement const& __e, int & __idGmshToFeel );
    void addVolume( mesh_type* /*mesh*/, Feel::detail::GMSHElement const& /*__e*/, int & /*__idGmshToFeel*/ , mpl::int_<1> );
    void addVolume( mesh_type* /*mesh*/, Feel::detail::GMSHElement const& /*__e*/, int & /*__idGmshToFeel*/ , mpl::int_<2> );
    void addVolume( mesh_type* mesh, Feel::detail::GMSHElement const& __e, int & /*__idGmshToFeel*/, mpl::int_<3> );

    void updateGhostCellInfo( mesh_type* mesh, std::vector<int> const& __idGmshToFeel, std::map<int,boost::tuple<int,int> > const& __mapGhostElt );


private:

    std::string _M_version;
    std::map<int,int> _M_n_vertices;
    //std::vector<int> _M_n_b_vertices;

    std::set<int> _M_ignorePhysicalGroup;
    std::set<std::string> _M_ignorePhysicalName;
    bool M_use_elementary_region_as_physical_region;

    //std::map<int,int> itoii;
    //std::vector<int> ptseen;

};



template<typename MeshType>
void
ImporterGmsh<MeshType>::showMe() const
{
    DVLOG(2) << "[ImporterGmsh::showMe] npoints_per_element = " << npoints_per_element << "\n";
    DVLOG(2) << "[ImporterGmsh::showMe]    npoints_per_face = " << npoints_per_face << "\n";
    DVLOG(2) << "[ImporterGmsh::showMe]    npoints_per_edge = " << npoints_per_edge << "\n";
}
template<typename MeshType>
boost::tuple<bool,boost::tuple<bool,int> >
ImporterGmsh<MeshType>::isElementOnProcessor( std::vector<int> const& tag ) const
{
    bool is_element_on_processor = false;

    // if there is no partition tag (only 2 tags) then consider all elements on
    // current processor
    if ( tag.size() == 2 )
        return boost::make_tuple( true,false );

    // tag[2] is the number of partition tags (1 in general 2 if cell share an
    // face with 2 processors)
    int rank = this->worldComm().localRank();
    auto it = std::find_if( tag.begin()+3, tag.end(), [&rank] ( int i )
    {
        return i == rank;
    } );

    if ( it != tag.end() )
        is_element_on_processor = true;

    bool isGhostCell=false;
    int idOfGhostCell=0;

    if ( tag[3]!=rank )
    {
        isGhostCell=true;
        idOfGhostCell=tag[3];
    }

    return boost::make_tuple( is_element_on_processor, boost::make_tuple( isGhostCell,idOfGhostCell ) );
}
template<typename MeshType>
void
ImporterGmsh<MeshType>::visit( mesh_type* mesh )
{

    if ( this->version() != "1.0" &&
         this->version() != "2.0" &&
         this->version() != "2.1" &&
         this->version() != "2.2" &&
         this->version() != FEELPP_GMSH_FORMAT_VERSION )
        throw std::logic_error( "invalid gmsh file format version" );

    DVLOG(2) << "visit("  << mesh_type::nDim << "D ) starts\n";
    DVLOG(2) << "visit("  << mesh_type::nDim << "D ) filename = " << this->filename() << "\n";

    std::ifstream __is ( this->filename().c_str() );

    if ( !__is.is_open() )
    {
        std::ostringstream ostr;
        LOG(ERROR) << "Invalid file name " << this->filename() << " (file not found)";
        ostr << "Invalid file name " << this->filename() << " (file not found)\n";
        throw std::invalid_argument( ostr.str() );
    }

    char __buf[256];
    __is >> __buf;

    std::string theversion;
    double version = 2.2;

    bool binary = false, swap = false;

    if ( ( ( this->version() == "2.0" ) ||
           ( this->version() == "2.1" ) ||
           ( this->version() == "2.2" ) ||
           ( this->version() == FEELPP_GMSH_FORMAT_VERSION ) )  &&
         std::string( __buf ) == "$MeshFormat" )
    {

        // version file-type(0=ASCII,1=BINARY) data-size(sizeof(double))
        int format, size;
        __is >> theversion >> format >> size;
        LOG(INFO) << "GMSH mesh file version : " << theversion << " format: " << (format?"binary":"ascii") << " size of double: " << size << "\n";
        CHECK( boost::lexical_cast<double>( theversion ) >= 2  )
            << "Feel++ supports only Gmsh version >= 2.\n";
        version =  boost::lexical_cast<double>( theversion );
        if(format)
        {
            char c;
            CHECK( (c=__is.get()) == '\n' ) << "Invalid character " << c << " should be newline\n";
            binary = true;
            LOG(INFO) << "GMSH mesh is in binary format\n";
            int one;
            __is.read( (char*)&one, sizeof(int) );
            if(one != 1)
            {
                swap = true;
                LOG(INFO) << "one before swap : " << one << "\n";
                if(swap) SwapBytes((char*)&one, sizeof(int), 1);
                LOG(INFO) << "one after swap : " << one << "\n";
                LOG(INFO) <<"Swapping bytes from binary file (to be done)\n";
            }
            //__is >> c;
            //LOG(INFO) << "character= " << c << "\n";
        }
        // should be $EndMeshFormat
        __is >> __buf;
        CHECK( std::string( __buf ) == "$EndMeshFormat" )
            << "invalid file format entry "
            << __buf
            << " instead of $EndMeshFormat\n";
        __is >> __buf;
        DVLOG(2) << "[importergmsh] " << __buf << " (expect $PhysicalNames)\n";

        std::vector<MeshMarkerName> meshMarkerNameMap = markerMap(MeshType::nDim);
        if ( std::string( __buf ) == "$PhysicalNames" )
        {
            int nnames;
            __is >> nnames;

            for ( int n = 0; n < nnames; ++n )
            {
                int id, topodim;
                std::string name;

                if ( boost::lexical_cast<double>( this->version() ) >= 2.1  )
                {
                    __is >> topodim >> id >> name;
                    DVLOG(2) << "[importergmsh] reading topodim: "  << topodim << " id: " << id << " name: " << name << "\n";
                }

                else if ( this->version() == "2.0" )
                    __is >> id >> name;

                boost::trim( name );
                boost::trim_if( name,boost::is_any_of( "\"" ) );

                if ( meshMarkerNameMap.empty() )
                {
                    std::vector<int> data = {id, topodim};
                    mesh->addMarkerName( name, id, topodim );
                }
                if ( _M_ignorePhysicalName.find( name )!=_M_ignorePhysicalName.end() ) this->setIgnorePhysicalGroup( id );
            }
            if ( meshMarkerNameMap.empty() )
            {
                FEELPP_ASSERT( mesh->markerNames().size() == ( size_type )nnames )( mesh->markerNames().size() )( nnames ).error( "invalid number of physical names" );
            }
            __is >> __buf;
            FEELPP_ASSERT( std::string( __buf ) == "$EndPhysicalNames" )
                ( __buf )
                ( "$EndPhysicalNames" ).error ( "invalid file format entry" );
            __is >> __buf;
        }

        for(auto it = meshMarkerNameMap.begin(), en = meshMarkerNameMap.end(); it != en; ++ it )
        {
            mesh->addMarkerName( it->name, it->ids[0], it->ids[1] );
        }
    }

    //
    // Read NODES
    //
    DVLOG(2) << "buf: "<< __buf << "\n";
    CHECK( std::string( __buf ) == "$NOD" ||
           std::string( __buf ) == "$Nodes" ||
           std::string( __buf ) == "$ParametricNodes" )
        << "invalid nodes string '" << __buf << "' in gmsh importer. It should be either $NOD or $Nodes or $ParametricNodes.\n";
    bool has_parametric_nodes = ( std::string( __buf ) == "$ParametricNodes" );
    uint __n;
    __is >> __n;

    // eat  '\n' in binary mode otherwise the next binary read will get screwd
    if ( binary )
        __is.get();


    std::map<int, Feel::detail::GMSHPoint > gmshpts;
    LOG(INFO) << "Reading "<< __n << " nodes\n";

    Eigen::Vector3d x;
    Eigen::Vector2d uv;
    for ( uint __i = 0; __i < __n; ++__i )
    {
        int id = 0;
        if ( !binary )
        {
            __is >> id
                 >> x[0]
                 >> x[1]
                 >> x[2];
        }
        else
        {
            __is.read( (char*)&id, sizeof(int) );
            if(swap) SwapBytes((char*)&id, sizeof(int), 1);
            __is.read( (char*)&x[0], 3*sizeof(double) );
            if(swap) SwapBytes((char*)&x[0], sizeof(double), 3);

        }
        gmshpts[id].id = id;
        gmshpts[id].x = x;

        if ( has_parametric_nodes )
        {

            gmshpts[id].parametric = true;
            CHECK( !binary ) << "GMSH Binary format not yet supported for parametric nodes\n";
            int gdim = 0, gtag = 0;
            if ( !binary )
            {
                __is >> gdim >> gtag;

                // if gdim == 0 then u = v = 0
                // if gdim == 3 then no need for a parametric point
                // this logic is done later when filling the mesh data structure
                if ( gdim == 1 )
                    __is >> uv[0];

                else if ( gdim == 2 )
                    __is >> uv[0] >> uv[1];
            }
            gmshpts[id].gdim = gdim;
            gmshpts[id].gtag = gtag;
            gmshpts[id].uv = uv;
        }

        // stores mapping to be able to reorder the indices
        // so that they are contiguous
        //itoii[idpts[__i]] = __i;
    }
    //ptseen.resize( __n );
    //std::fill( ptseen.begin(), ptseen.end(), -1 );
    // eat  '\n' in binary mode otherwise the next binary read will get screwd
    if ( binary )
        __is.get();

    __is >> __buf;
    DVLOG(2) << "buf: "<< __buf << "\n";
    // make sure that we have read all the points
    CHECK( std::string( __buf ) == "$ENDNOD" ||
           std::string( __buf ) == "$EndNodes" ||
           std::string( __buf ) == "$EndParametricNodes" )
        << "invalid end nodes string '" << __buf
        << "' in gmsh importer. It should be either $ENDNOD or $EndNodes or $EndParametricNodes\n";

    //
    // Read ELEMENTS
    //
    __is >> __buf;
    CHECK( std::string( __buf ) == "$ELM" ||
           std::string( __buf ) == "$Elements" )
        << "invalid elements string " << __buf << " in gmsh importer, it should be either $ELM or $Elements\n";

    int numElements;
    __is >> numElements;

    // eat  '\n' in binary mode otherwise the next binary read will get screwd
    if ( binary )
        __is.get();

    LOG(INFO) << "Reading " << numElements << " elements...\n";
    std::vector<Feel::detail::GMSHElement> __et; // tags in each element
    std::vector<int> __idGmshToFeel( numElements ); // id Gmsh to id Feel
    std::map<int,int> __gt;

    if ( !binary )
    {
        for(int i = 0; i < numElements; i++)
        {
          int num, type, physical = 0, elementary = 0, parent = 0;
          int dom1 = 0, dom2 = 0, numVertices;
          std::vector<int> ghosts;
          int numTags;
          // some faces may not be associated to a partition in the mesh file,
          // hence will be read given the partition id 0 and will be discarded
          int partition = (this->worldComm().globalSize()>1)?this->worldComm().localRank():0;
          __is >> num  // elm-number
               >> type // elm-type
               >> numTags; // number-of-tags

          int numPartitions = 1;

          for(int j = 0; j < numTags; j++)
          {
              int tag;
              __is >> tag;
              if(j == 0) physical = tag;
              else if(j == 1) elementary = tag;
              else if(version < 2.2 && j == 2) partition = tag;
              else if(version >= 2.2 && j == 2 && numTags > 3) numPartitions = tag;
              else if(version >= 2.2 && j == 3) partition = tag-1;
              else if(j >= 4 && j < 4 + numPartitions - 1) ghosts.push_back((-tag)-1);
              else if(j == 3 + numPartitions && (numTags == 4 + numPartitions))
                  parent = tag;
              else if(j == 3 + numPartitions && (numTags == 5 + numPartitions)) {
                  dom1 = tag; j++;
                  __is >> dom2;
              }
          }
          CHECK(type != MSH_POLYG_ && type != MSH_POLYH_ && type != MSH_POLYG_B)
              << "GMSH Element type " << type << " not supported by Feel++\n";
          numVertices = MElement::getInfoMSH(type);
          CHECK(numVertices!=0) << "Unknown number of vertices for element type " << type << "\n";

          std::vector<int> indices(numVertices);
          for(int j = 0; j < numVertices; j++)
          {
              __is >> indices[j];
              // we do not renumber anymore
#if 0
              indices[j] = itoii[ indices[j] ];
#endif
          }
          if ( M_use_elementary_region_as_physical_region )
          {
              physical = elementary;
          }

          __et.push_back( Feel::detail::GMSHElement( num, type, physical, elementary,
                                                     numPartitions, partition, ghosts,
                                                     parent, dom1, dom2,
                                                     numVertices, indices ) );
          if ( __gt.find( type ) != __gt.end() )
              ++__gt[ type ];
          else
              __gt[type]=1;


        }  // element description loop
    } // !binary
    else // binary case
    {
        int numElementsPartial = 0;
        while(numElementsPartial < numElements)
        {
            int header[3];

            __is.read( (char*)&header, 3*sizeof(int) );
            if(swap) SwapBytes((char*)header, sizeof(int), 3);

            int type = header[0];
            int numElems = header[1];
            int numTags = header[2];
            char const* name;
            CHECK( type < MSH_NUM_TYPE ) << "Invalid GMSH element type " << type << "\n";
            int numVertices = MElement::getInfoMSH(type,&name);
            CHECK( numVertices > 0 ) << "Unsupported element type " << name << "\n";

            unsigned int n = 1 + numTags + numVertices;
            std::vector<int> data(n);
            std::vector<int> indices( numVertices );
            std::vector<int> ghosts;

            for(int i = 0; i < numElems; i++)
            {
                ghosts.clear();
                __is.read( (char*)data.data(), sizeof(int)*n );
                if(swap) SwapBytes((char*)data.data(), sizeof(int), n);

                //if(swap) SwapBytes((char*)data, sizeof(int), n);
                int num = data[0];
                int physical = (numTags > 0) ? data[1] : 0;
                int elementary = (numTags > 1) ? data[2] : 0;
                int numPartitions = (version >= 2.2 && numTags > 3) ? data[3] : 1;
                int partition = (version < 2.2 && numTags > 2) ? data[3]-1 :
                    (version >= 2.2 && numTags > 3) ? data[4]-1 : 0;
                if(numPartitions > 1)
                    for(int j = 0; j < numPartitions - 1; j++)
                        ghosts.push_back( (-data[5 + j]) -1 );
                int parent = (version < 2.2 && numTags > 3) ||
                    (version >= 2.2 && numPartitions && numTags > 3 + numPartitions) ||
                    (version >= 2.2 && !numPartitions && numTags > 2) ?
                    data[numTags] : 0;
                int dom1 = 0, dom2 = 0;

                std::copy( &data[numTags + 1], &data[numTags + 1]+numVertices, indices.begin() );

                // we do not renumber anymore
#if 0
                for(int j = 0; j < numVertices; j++)
                {
                    indices[j] = itoii[ indices[j] ];
                }
#endif
                if ( M_use_elementary_region_as_physical_region )
                {
                    physical = elementary;
                }

                __et.push_back( Feel::detail::GMSHElement( num, type, physical, elementary,
                                                           numPartitions, partition, ghosts,
                                                           parent, dom1, dom2,
                                                           numVertices, indices ) );

                if ( __gt.find( type ) != __gt.end() )
                    ++__gt[ type ];
                else
                    __gt[type]=1;

            }
            numElementsPartial += numElems;

        } // while
        CHECK( numElementsPartial == numElements ) << "Invalid number of elements read from GMSH, read " << numElementsPartial << " element but expected " << numElements << "\n";
    } // binary

    auto it = __gt.begin();
    auto en = __gt.end();
    for( ; it != en; ++it )
    {
        const char* name;
        MElement::getInfoMSH( it->first, &name );
        LOG(INFO) << "Read " << it->second << " " << name << " elements\n";
    }

    if ( binary )
        __is >> __buf;

    // make sure that we have read everything
    __is >> __buf;
    CHECK( std::string( __buf ) == "$ENDELM" ||
           std::string( __buf ) == "$EndElements" )
        << "invalid end elements string " << __buf
        << " in gmsh importer. It should be either $ENDELM or $EndElements\n";

#if 0
    //
    // FILL Mesh Data Structure
    //
    for ( uint __i = 0; __i < numElements; ++__i )
    {
        // if the element is not associated to the processor (in partition or ghost) or
        // if the physical entity is ignored
        if ( __et[__i].isOnProcessor() == false ||
             __et[__i].isIgnored(_M_ignorePhysicalGroup.begin(), _M_ignorePhysicalGroup.end()) )
            continue;

        switch ( __et[__i].type )
        {
        case GMSH_POINT:
            if ( mesh_type::nDim == 1 )
            {
                gmshpts[ __et[__i].indices[0] ].onbdy = true;
            }

            break;

        case GMSH_LINE:
        case GMSH_LINE_2:
        case GMSH_LINE_3:
        case GMSH_LINE_4:
        case GMSH_LINE_5:
            if ( mesh_type::nDim == 2 )
            {
                for ( uint16_type jj = 0; jj < npoints_per_edge; ++jj )
                {
                    gmshpts[ __et[__i].indices[jj] ].onbdy = true;
                }
            }

            break;

        case GMSH_QUADRANGLE:
        case GMSH_QUADRANGLE_2:
        case GMSH_TRIANGLE:
        case GMSH_TRIANGLE_2:
        case GMSH_TRIANGLE_3:
        case GMSH_TRIANGLE_4:
        case GMSH_TRIANGLE_5:
            if ( mesh_type::nDim == 3 )
            {
                for ( uint16_type jj = 0; jj < npoints_per_face; ++jj )
                {
                    gmshpts[ __et[__i].indices[jj] ].onbdy = true;
                }
            }

            break;

        default:
            break;
        }
    }
#endif
    std::map<int,boost::tuple<int,int> > mapGhostElt;

    node_type coords( mesh_type::nRealDim );

    //_M_n_b_vertices.resize( __n );
    //_M_n_b_vertices.assign( __n, 0 );

    // add the element to the mesh
    for ( int __i = 0; __i < numElements; ++__i )
    {
        // if the element is not associated to the processor (in partition or ghost) or
        // if the physical entity is ignored
        if ( __et[__i].isOnProcessor() == false ||
             __et[__i].isIgnored(_M_ignorePhysicalGroup.begin(), _M_ignorePhysicalGroup.end()) )
            continue;
        // add the points associates to the element on the processor
        for ( uint16_type p = 0; p < __et[__i].numVertices; ++p )
        {
            int ptid = __et[__i].indices[p];
            // don't do anything if the point is already registered
            if ( mesh->hasPoint( ptid ) )
                continue;

            auto const& gmshpt = gmshpts.find(ptid)->second;
            for ( uint16_type j = 0; j < mesh_type::nRealDim; ++j )
                coords[j] = gmshpt.x[j];

            point_type pt( ptid, coords, gmshpt.onbdy );

            if ( gmshpt.parametric )
            {
                pt.setGDim( gmshpt.gdim );
                pt.setGTag( gmshpt.gtag );

                if ( gmshpt.gdim < 3 )
                {
                    pt.setParametricCoordinates( gmshpt.uv[0], gmshpt.uv[1] );
                    mesh->setParametric( true );
                }
            }
            mesh->addPoint( pt );
        } // loop over local points

        switch ( __et[__i].type )
        {
            // Points
        case GMSH_POINT:
        {
            addPoint( mesh, __et[__i], __idGmshToFeel[__i] );


            break;
        }

        // Edges
        case GMSH_LINE:
        case GMSH_LINE_2:
        case GMSH_LINE_3:
        case GMSH_LINE_4:
        case GMSH_LINE_5:
        {
            addEdge( mesh, __et[__i], __idGmshToFeel[__i] );

            break;
        }

        // Faces
        case GMSH_TRIANGLE:
        case GMSH_TRIANGLE_2:
        case GMSH_TRIANGLE_3:
        case GMSH_TRIANGLE_4:
        case GMSH_TRIANGLE_5:
        case GMSH_QUADRANGLE:
        case GMSH_QUADRANGLE_2:
        {
            addFace( mesh, __et[__i], __idGmshToFeel[__i] );

            break;
        }

        // Volumes
        case GMSH_TETRAHEDRON:
        case GMSH_TETRAHEDRON_2:
        case GMSH_TETRAHEDRON_3:
        case GMSH_TETRAHEDRON_4:
        case GMSH_TETRAHEDRON_5:
        case GMSH_HEXAHEDRON:
        case GMSH_HEXAHEDRON_2:
        {
            addVolume( mesh, __et[__i], __idGmshToFeel[__i] );

            break;
        }

        default:
            break;
        }

        if ( __et[__i].isGhost() )
        {
            mapGhostElt.insert( std::make_pair( __i,boost::make_tuple( __idGmshToFeel[__i], __et[__i].ghostPartitionId() ) ) );
        }

    } // loop over geometric entities in gmsh file (can be elements or faces)

    if (VLOG_IS_ON(4))
    {
        //for(int i = 0; i < ptseen.size();  ++i )
        //if ( ptseen[i] == -1 )
        //LOG(WARNING) << "Point with id " << i << " not in element connectivity";
    }
    CHECK( mesh->numElements() > 0 ) << "The mesh does not have any elements.\n"
                                     << "something was not right with GMSH mesh importation.\n"
                                     << "please check that there are elements of topological dimension "
                                     << mesh_type::nDim << "  in the mesh\n";

    if ( this->worldComm().localSize()>1 )
        updateGhostCellInfo( mesh, __idGmshToFeel,  mapGhostElt );

    mesh->setNumVertices( std::accumulate( _M_n_vertices.begin(), _M_n_vertices.end(), 0,
                                           []( int lhs, std::pair<int,int> const& rhs )
                                           {
                                               return lhs+rhs.second;
                                           } ) );
    if ( !mesh->markerNames().empty() &&
         ( mesh->markerNames().find("CrossPoints") != mesh->markerNames().end() ) &&
         ( mesh->markerNames().find("WireBasket") != mesh->markerNames().end() ) )
    {
        //LOG(INFO) << "[substructuring] marker cp" << mesh->markerName("CrossPoints")  << "\n";
        //LOG(INFO) << "[substructuring] marker wb" << mesh->markerName("WireBasket")  << "\n";
        //LOG(INFO) << "[substructuring] n cp: " << std::distance( mesh->beginPointWithMarker( mesh->markerName("CrossPoints") ), mesh->endPointWithMarker( mesh->markerName("CrossPoints") ) ) << "\n";
    }
    DVLOG(2) << "done with reading and creating mesh from gmsh file\n";
    _M_n_vertices.clear();
}

template<typename MeshType>
void
ImporterGmsh<MeshType>::addPoint( mesh_type* mesh, Feel::detail::GMSHElement const& __e, int & __idGmshToFeel )
{
    addPoint( mesh, __e, __idGmshToFeel, mpl::int_<mesh_type::nDim>() );
}
template<typename MeshType>
void
ImporterGmsh<MeshType>::addPoint( mesh_type*mesh, Feel::detail::GMSHElement const& __e, int & __idGmshToFeel, mpl::int_<1> )
{
    face_type pf;
    pf.setProcessIdInPartition( this->worldComm().localRank() );
    pf.setId( mesh->numFaces() );
    pf.setMarker( __e.physical );
    pf.setMarker2( __e.elementary );
    pf.setNumberOfPartitions( __e.numPartitions );
    pf.setProcessId( __e.partition );
    pf.setNeighborPartitionIds( __e.ghosts );

    pf.setPoint( 0, mesh->point( __e.indices[0] ) );
//    ptseen[mesh->point( __e.indices[0] ).id()]=1;
    if ( mesh->point( __e.indices[0] ).isOnBoundary() )
    {
        pf.setOnBoundary( true );
    }
    _M_n_vertices[ __e.indices[0] ] = 1;

    //_M_n_b_vertices[ __e.indices[0] ] = 1;

    face_iterator fit;
    bool inserted;
    boost::tie( fit, inserted ) = mesh->addFace( pf );
    __idGmshToFeel=pf.id();

    DVLOG(2) << "added point on boundary ("
                  << fit->isOnBoundary() << ") with id :" << fit->id() << " and marker " << pf.marker()
                  << " n1: " << mesh->point( __e.indices[0] ).node() << "\n";
}
template<typename MeshType>
void
ImporterGmsh<MeshType>::addPoint( mesh_type* mesh, Feel::detail::GMSHElement const& __e, int & __idGmshToFeel, mpl::int_<2> )
{
    auto pit = mesh->points().modify( mesh->pointIterator(__e.indices[0]),
                                      [&__e]( point_type& e )
                                      {
                                          e.setMarker( __e.physical );
                                          e.setMarker2( __e.elementary );
                                          e.setNumberOfPartitions( __e.numPartitions );
                                          e.setProcessId( __e.partition );
                                          e.setNeighborPartitionIds( __e.ghosts );
                                      } );
    DVLOG(2) << "added point with id :" << mesh->pointIterator(__e.indices[0])->id() << " and marker " << mesh->pointIterator(__e.indices[0])->marker()
                  << " n1: " << mesh->point( __e.indices[0] ).node() << "\n";

}
template<typename MeshType>
void
ImporterGmsh<MeshType>::addPoint( mesh_type* mesh, Feel::detail::GMSHElement const& __e, int & __idGmshToFeel, mpl::int_<3> )
{
    auto pit = mesh->pointIterator(__e.indices[0]);
    bool mod = mesh->points().modify( pit,
                                      [&__e]( point_type& e )
                                      {
                                          e.setMarker( __e.physical );
                                          e.setMarker2( __e.elementary );
                                          e.setNumberOfPartitions( __e.numPartitions );
                                          e.setProcessId( __e.partition );
                                          e.setNeighborPartitionIds( __e.ghosts );
                                      } );
    DVLOG(2) << "added point (modified: " << mod << ")with id :" << pit->id() << " and marker " << pit->marker()
                  << " n1: " << mesh->point( __e.indices[0] ).node() << "\n";
}

template<typename MeshType>
void
ImporterGmsh<MeshType>::addEdge( mesh_type* mesh, Feel::detail::GMSHElement const& __e, int & __idGmshToFeel )
{
    addEdge( mesh, __e, __idGmshToFeel, mpl::int_<mesh_type::nDim>() );
}
template<typename MeshType>
void
ImporterGmsh<MeshType>::addEdge( mesh_type*mesh, Feel::detail::GMSHElement const& __e, int & __idGmshToFeel, mpl::int_<1> )
{
    element_type e;
    //e.setWorldComm(this->worldComm());
    e.setProcessIdInPartition( this->worldComm().localRank() );
    e.setMarker( __e.physical );
    e.setMarker2( __e.elementary );
    e.setNumberOfPartitions( __e.numPartitions );
    e.setProcessId( __e.partition );
    e.setNeighborPartitionIds( __e.ghosts );

    if ( __e.type == GMSH_LINE ||
         __e.type == GMSH_LINE_2 ||
         __e.type == GMSH_LINE_3 ||
         __e.type == GMSH_LINE_4 ||
         __e.type == GMSH_LINE_5 )
    {
        int count_pt_on_boundary = 0;
        for ( uint16_type jj = 0; jj < npoints_per_element; ++jj )
        {
            e.setPoint( jj, mesh->point( __e.indices[jj] ) );
            //ptseen[mesh->point( __e.indices[jj] ).id()]=1;
            if ( mesh->point( __e.indices[jj] ).isOnBoundary() )
                ++count_pt_on_boundary;
        }
        if ( count_pt_on_boundary >= 1 )
        {
            e.setOnBoundary( true );
        }

    }

    mesh->addElement( e );
    __idGmshToFeel=e.id();

    _M_n_vertices[ __e.indices[0] ] = 1;
    _M_n_vertices[ __e.indices[1] ] = 1;
    DVLOG(2) << "added edge with id :" << e.id()
                  << " n1: " << mesh->point( __e.indices[0] ).node()
                  << " n2: " << mesh->point( __e.indices[1] ).node() << "\n";
}

template<typename MeshType>
void
ImporterGmsh<MeshType>::addEdge( mesh_type* mesh, Feel::detail::GMSHElement const& __e, int & __idGmshToFeel, mpl::int_<2> )
{
    face_type e;
    e.setProcessIdInPartition( this->worldComm().localRank() );
    e.setId( mesh->numFaces() );
    e.setMarker( __e.physical );
    e.setMarker2( __e.elementary );
    e.setNumberOfPartitions( __e.numPartitions );
    e.setProcessId( __e.partition );
    e.setNeighborPartitionIds( __e.ghosts );

    if ( __e.type == GMSH_LINE ||
         __e.type == GMSH_LINE_2 ||
         __e.type == GMSH_LINE_3 ||
         __e.type == GMSH_LINE_4 ||
         __e.type == GMSH_LINE_5 )
    {
        int count_pt_on_boundary = 0;
        for ( uint16_type jj = 0; jj < npoints_per_edge; ++jj )
        {
            e.setPoint( jj, mesh->point( __e.indices[jj] ) );
            if ( mesh->point( __e.indices[jj] ).isOnBoundary() )
                ++count_pt_on_boundary;
        }
        if ( count_pt_on_boundary >= 2 )
            e.setOnBoundary( true );

    }

    _M_n_vertices[ __e.indices[0] ] = 1;
    _M_n_vertices[ __e.indices[1] ] = 1;

    //_M_n_b_vertices[ __e.indices[0] ] = 1;
    //_M_n_b_vertices[ __e.indices[1] ] = 1;

    bool inserted;
    face_iterator fit;
    boost::tie( fit, inserted ) = mesh->addFace( e );
    __idGmshToFeel=e.id();

    DVLOG(2) << "added edge on boundary ("
                  << fit->isOnBoundary() << ") with id :" << fit->id()
                  << " n1: " << mesh->point( __e.indices[0] ).node()
                  << " n2: " << mesh->point( __e.indices[1] ).node() << "\n";
}
template<typename MeshType>
void
ImporterGmsh<MeshType>::addEdge( mesh_type*mesh, Feel::detail::GMSHElement const& __e, int & __idGmshToFeel, mpl::int_<3> )
{
    edge_type e;
    e.setProcessIdInPartition( this->worldComm().localRank() );
    e.setId( mesh->numEdges() );
    e.setMarker( __e.physical );
    e.setMarker2( __e.elementary );
    e.setNumberOfPartitions( __e.numPartitions );
    e.setProcessId( __e.partition );
    e.setNeighborPartitionIds( __e.ghosts );

    if ( __e.type == GMSH_LINE ||
         __e.type == GMSH_LINE_2 ||
         __e.type == GMSH_LINE_3 ||
         __e.type == GMSH_LINE_4 ||
         __e.type == GMSH_LINE_5 )
    {
        int count_pt_on_boundary = 0;
        for ( uint16_type jj = 0; jj < npoints_per_edge; ++jj )
        {
            e.setPoint( jj, mesh->point( __e.indices[jj] ) );
            if ( mesh->point( __e.indices[jj] ).isOnBoundary() )
                ++count_pt_on_boundary;
        }
        if ( count_pt_on_boundary >= 2 )
        {
            e.setOnBoundary( true );
        }
    }

    _M_n_vertices[ __e.indices[0] ] = 1;
    _M_n_vertices[ __e.indices[1] ] = 1;

    //_M_n_b_vertices[ __e.indices[0] ] = 1;
    //_M_n_b_vertices[ __e.indices[1] ] = 1;

    auto eit = mesh->addEdge( e );
    __idGmshToFeel=eit.id();

    if ( npoints_per_edge == 2 )
        DVLOG(2) << "added edge on boundary ("
                      << eit.isOnBoundary() << ") with id :" << eit.id()
                      << " n1: " << eit.point( 0 ).node()
                      << " n2: " << eit.point( 1 ).node() << "\n";

}

template<typename MeshType>
void
ImporterGmsh<MeshType>::addFace( mesh_type* mesh, Feel::detail::GMSHElement const& __e, int & __idGmshToFeel )
{
    addFace( mesh, __e, __idGmshToFeel, mpl::int_<mesh_type::nDim>() );
}

template<typename MeshType>
void
ImporterGmsh<MeshType>::addFace( mesh_type*, Feel::detail::GMSHElement const&, int & /*__idGmshToFeel*/, mpl::int_<1> )
{}

template<typename MeshType>
void
ImporterGmsh<MeshType>::addFace( mesh_type* mesh, Feel::detail::GMSHElement const& __e, int & __idGmshToFeel, mpl::int_<2> )
{
    GmshOrdering<element_type> ordering;

    element_type e;
    e.setProcessIdInPartition( this->worldComm().localRank() );
    e.setMarker( __e.physical );
    e.setMarker2( __e.elementary );
    e.setNumberOfPartitions( __e.numPartitions );
    e.setProcessId( __e.partition );
    e.setNeighborPartitionIds( __e.ghosts );

    if ( __e.type == GMSH_QUADRANGLE ||
         __e.type == GMSH_QUADRANGLE_2 ||
         __e.type == GMSH_TRIANGLE ||
         __e.type == GMSH_TRIANGLE_2 ||
         __e.type == GMSH_TRIANGLE_3 ||
         __e.type == GMSH_TRIANGLE_4 ||
         __e.type == GMSH_TRIANGLE_5 )
    {
        int count_pt_on_boundary = 0;
        for ( uint16_type jj = 0; jj < npoints_per_element; ++jj )
        {
            //ptseen[mesh->point( __e.indices[jj] ).id()]=1;
            if (!e.isGhostCell()) mesh->points().modify( mesh->pointIterator( __e.indices[jj] ), Feel::detail::UpdateProcessId(e.processId()) );
            e.setPoint( ordering.fromGmshId( jj ), mesh->point( __e.indices[jj] ) );
            if ( mesh->point( __e.indices[jj] ).isOnBoundary() )
                ++count_pt_on_boundary;
        }
        if ( ( __e.type == GMSH_TRIANGLE ||
               __e.type == GMSH_TRIANGLE_2 ||
               __e.type == GMSH_TRIANGLE_3 ||
               __e.type == GMSH_TRIANGLE_4 ||
               __e.type == GMSH_TRIANGLE_5 ) &&
             count_pt_on_boundary >= 2 )
        {
            e.setOnBoundary( true );
        }
        if ( (__e.type == GMSH_QUADRANGLE ||
              __e.type == GMSH_QUADRANGLE_2) &&
             count_pt_on_boundary >= 2 )
            e.setOnBoundary( true );

    }

    mesh->addElement( e );
    __idGmshToFeel=e.id();

    _M_n_vertices[ __e.indices[0] ] = 1;
    _M_n_vertices[ __e.indices[1] ] = 1;
    _M_n_vertices[ __e.indices[2] ] = 1;

    if ( __e.type == GMSH_QUADRANGLE ||
         __e.type == GMSH_QUADRANGLE_2 )
        _M_n_vertices[ __e.indices[3] ] = 1;
}
template<typename MeshType>
void
ImporterGmsh<MeshType>::addFace( mesh_type* mesh, Feel::detail::GMSHElement const& __e, int & __idGmshToFeel, mpl::int_<3> )
{
    GmshOrdering<face_type> ordering;

    face_type e;
    e.setProcessIdInPartition( this->worldComm().localRank() );
    e.setId( mesh->numFaces() );
    e.setMarker( __e.physical );
    e.setMarker2( __e.elementary );
    e.setNumberOfPartitions( __e.numPartitions );
    e.setProcessId( __e.partition );
    e.setNeighborPartitionIds( __e.ghosts );

    // we consider in 3D that all faces provided by Gmsh are on the boundary
    // which might not be true
    // \warning there might be a bug here
    e.setOnBoundary( true );

    if ( __e.type == GMSH_QUADRANGLE ||
         __e.type == GMSH_QUADRANGLE_2 ||
         __e.type == GMSH_TRIANGLE ||
         __e.type == GMSH_TRIANGLE_2 ||
         __e.type == GMSH_TRIANGLE_3 ||
         __e.type == GMSH_TRIANGLE_4 ||
         __e.type == GMSH_TRIANGLE_5 )
    {
        for ( uint16_type jj = 0; jj < npoints_per_face; ++jj )
        {
            //ptseen[mesh->point( __e.indices[jj] ).id()]=1;
            e.setPoint( ordering.fromGmshId( jj ), mesh->point( __e.indices[jj] ) );
        }

        //e.setPoint( jj, mesh->point( __e[jj] ) );
    }

    bool inserted;
    face_iterator fit;
    boost::tie( fit, inserted ) = mesh->addFace( e );

    __idGmshToFeel=e.id();

    _M_n_vertices[ __e.indices[0] ] = 1;
    _M_n_vertices[ __e.indices[1] ] = 1;
    _M_n_vertices[ __e.indices[2] ] = 1;

    if ( __e.type == GMSH_QUADRANGLE ||
         __e.type == GMSH_QUADRANGLE_2 )
        _M_n_vertices[ __e.indices[3] ] = 1;

}

template<typename MeshType>
void
ImporterGmsh<MeshType>::addVolume( mesh_type* mesh, Feel::detail::GMSHElement const& __e, int & __idGmshToFeel )
{
    addVolume( mesh, __e, __idGmshToFeel, mpl::int_<mesh_type::nDim>() );
}
template<typename MeshType>
void
ImporterGmsh<MeshType>::addVolume( mesh_type*, Feel::detail::GMSHElement const&, int & /*__idGmshToFeel*/, mpl::int_<1> )
{}
template<typename MeshType>
void
ImporterGmsh<MeshType>::addVolume( mesh_type*, Feel::detail::GMSHElement const&, int & /*__idGmshToFeel*/, mpl::int_<2> )
{}
template<typename MeshType>
void
ImporterGmsh<MeshType>::addVolume( mesh_type* mesh, Feel::detail::GMSHElement const& __e, int & __idGmshToFeel, mpl::int_<3> )
{
    element_type e;
    e.setProcessIdInPartition( this->worldComm().localRank() );
    GmshOrdering<element_type> ordering;
    e.setMarker( __e.physical );
    e.setMarker2( __e.elementary );
    e.setNumberOfPartitions( __e.numPartitions );
    e.setProcessId( __e.partition );
    e.setNeighborPartitionIds( __e.ghosts );


    //
    // WARNING: not yet done for high order in 3D !!!
    //
    if ( __e.type == GMSH_HEXAHEDRON ||
         __e.type == GMSH_HEXAHEDRON_2 ||
         __e.type == GMSH_TETRAHEDRON ||
         __e.type == GMSH_TETRAHEDRON_2 ||
         __e.type == GMSH_TETRAHEDRON_3 ||
         __e.type == GMSH_TETRAHEDRON_4 ||
         __e.type == GMSH_TETRAHEDRON_5 )
    {
        int count_pt_on_boundary = 0;
        for ( uint16_type jj = 0; jj < npoints_per_element; ++jj )
        {
            //ptseen[mesh->point( __e.indices[jj] ).id()]=1;
            if (!e.isGhostCell()) mesh->points().modify( mesh->pointIterator( __e.indices[jj] ), Feel::detail::UpdateProcessId(e.processId()) );
            //std::cout << "gmsh index " << jj << " -> " << ordering.fromGmshId(jj) << " -> " << mesh->point( __e[jj] ).id()+1 << " : " << mesh->point( __e[jj] ).node() << "\n";
            e.setPoint( ordering.fromGmshId( jj ), mesh->point( __e.indices[jj] ) );

            if ( mesh->point( __e.indices[jj] ).isOnBoundary() )
                ++count_pt_on_boundary;
        }
        // the tet share a face with the boundary
        if ( ( __e.type == GMSH_TETRAHEDRON ||
               __e.type == GMSH_TETRAHEDRON_2 ||
               __e.type == GMSH_TETRAHEDRON_3 ||
               __e.type == GMSH_TETRAHEDRON_4 ||
               __e.type == GMSH_TETRAHEDRON_5 )
             && count_pt_on_boundary >= 3)
            e.setOnBoundary( true );
        // the hex share a face with the boundary
        if ( ( __e.type == GMSH_HEXAHEDRON ||
               __e.type == GMSH_HEXAHEDRON_2 )
             && count_pt_on_boundary >= 4)
            e.setOnBoundary( true );
    }

    mesh->addElement( e );
    __idGmshToFeel=e.id();

    _M_n_vertices[ __e.indices[0] ] = 1;
    _M_n_vertices[ __e.indices[1] ] = 1;
    _M_n_vertices[ __e.indices[2] ] = 1;
    _M_n_vertices[ __e.indices[3] ] = 1;

    if ( __e.type == GMSH_HEXAHEDRON )
    {
        _M_n_vertices[ __e.indices[4] ] = 1;
        _M_n_vertices[ __e.indices[5] ] = 1;
        _M_n_vertices[ __e.indices[6] ] = 1;
        _M_n_vertices[ __e.indices[7] ] = 1;
    }
}

template<typename MeshType>
void
ImporterGmsh<MeshType>::updateGhostCellInfo( mesh_type* mesh, std::vector<int> const& __idGmshToFeel, std::map<int,boost::tuple<int,int> > const& __mapGhostElt )
{
    // counter of msg sent for each process
    std::vector<int> nbMsgToSend( this->worldComm().localSize() );
    std::fill( nbMsgToSend.begin(),nbMsgToSend.end(),0 );

    // map usefull to get final result
    std::vector< std::map<int,int> > mapMsg( this->worldComm().localSize() );

    // iterate over ghost elt
    auto it_map = __mapGhostElt.begin();
    auto en_map = __mapGhostElt.end();

    for ( int cpt=0; it_map!=en_map; ++it_map,++cpt )
    {
        auto idGmsh = it_map->first;
        auto idProc = it_map->second.template get<1>();
#if 0
        std::cout << "[updateGhostCellInfo]----1---\n"
                  << "I am the proc " << this->worldComm().globalRank()
                  << " local proc " << this->worldComm().localRank()
                  << " I send to the proc " << idProc << " for idGmsh " << idGmsh+1
                  << " with tag "<< nbMsgToSend[idProc]
                  << " the G " << mesh->element( it_map->second.template get<0>(),idProc ).G()
                  << std::endl;
#endif
        // send
        this->worldComm().localComm().send( idProc , nbMsgToSend[idProc], idGmsh );
        // save tag of request
        mapMsg[idProc].insert( std::make_pair( nbMsgToSend[idProc],it_map->second.template get<0>() ) );
        // update nb send
        nbMsgToSend[idProc]++;
    }

    // counter of msg received for each process
    std::vector<int> nbMsgToRecv;
    mpi::all_to_all( this->worldComm().localComm(),
                     nbMsgToSend,
                     nbMsgToRecv );

    // get gmsh id asked and re-send the correspond id Feel
    for ( int proc=0; proc<this->worldComm().localSize(); ++proc )
    {
        for ( int cpt=0 ; cpt<nbMsgToRecv[proc] ; ++cpt )
        {
            int idGmsh;
            //reception idGmsh
            this->worldComm().localComm().recv( proc, cpt, idGmsh );
#if 0
            std::cout << "[updateGhostCellInfo]----2---\n"
                      << "I am the proc" << this->worldComm().localRank()
                      << " I receive of the proc " << proc << " for idGmsh " << idGmsh+1
                      << " with tag "<< cpt
                      << " the G " << mesh->element( __idGmshToFeel[idGmsh] ).G()
                      << " with idFeel Classic " << __idGmshToFeel[idGmsh]
                      << std::endl;
#endif
            //re-send idFeel
            this->worldComm().localComm().send( proc, cpt, __idGmshToFeel[ idGmsh ] );
        }
    }

    // get response to initial request and update Feel::Mesh data
    for ( int proc=0; proc<this->worldComm().localSize(); ++proc )
    {
        for ( int cpt=0; cpt<nbMsgToSend[proc]; ++cpt )
        {
            int idFeel;
            // receive idFeel
            this->worldComm().localComm().recv( proc, cpt, idFeel );
            // update data
            auto elttt = mesh->elementIterator( mapMsg[proc][cpt],proc );
            mesh->elements().modify( elttt, Feel::detail::updateIdInOthersPartitions( proc, idFeel ) );
#if 0
            std::cout << "[updateGhostCellInfo]----3---\n"
                      << "END! I am the proc" << this->worldComm().localRank()<<" I receive of the proc " << proc
                      <<" with tag "<< cpt
                      << " for the G " << mesh->element( mapMsg[proc][cpt], proc ).G()
                      << " with idFeel Classic " << mapMsg[proc][cpt]
                      << " with idFeel " << idFeel
                      << " and modif " << mesh->element( mapMsg[proc][cpt] , proc ).idInPartition( proc )
                      << std::endl;
#endif
        }
    }


    //check
#if 0
    std::vector<int> nbMsgToSendCheck( this->worldComm().localSize() );
    std::fill( nbMsgToSendCheck.begin(),nbMsgToSendCheck.end(),0 );

    std::vector< std::map<int,int> > mapMsgCheck( this->worldComm().localSize() );
    auto it_ghost=mesh->beginGhostElement();
    auto en_ghost=mesh->endGhostElement();

    for ( int cpt=0 ; it_ghost != en_ghost; ++it_ghost,++cpt )
    {

        //ATTENTION WARNING int OK, but auto no!!!!!!!!!!!!!!!!!!!!!!!
        int/*auto*/ IdProcessOfGhost = it_ghost->processId();
        int/*auto*/ RealIdOfGhost = it_ghost->idInPartition( IdProcessOfGhost );
        this->worldComm().localComm().send( IdProcessOfGhost, nbMsgToSendCheck[IdProcessOfGhost], RealIdOfGhost );
#if 0
        std::cout<< "I am the proc" << this->worldComm().localRank() << " I send to the proc " << IdProcessOfGhost
                 <<" with tag "<< nbMsgToSendCheck[IdProcessOfGhost]
                 << " idSend " << RealIdOfGhost
                 << " it_ghost->G() " << it_ghost->G()
                 << std::endl;
#endif
        mapMsgCheck[IdProcessOfGhost].insert( std::make_pair( cpt,nbMsgToSendCheck[IdProcessOfGhost] ) );

        ++nbMsgToSendCheck[IdProcessOfGhost];
    }

    // counter of msg received for each process
    std::vector<int> nbMsgToRecvCheck;
    mpi::all_to_all( this->worldComm().localComm(),
                     nbMsgToSendCheck,
                     nbMsgToRecvCheck );

    for ( int proc=0; proc<this->worldComm().localSize(); ++proc )
    {
        for ( int cpt=0; cpt<nbMsgToRecvCheck[proc]; ++cpt )
        {
            int idRecv;
            this->worldComm().localComm().recv( proc, cpt,idRecv );
#if 0
            std::cout<< "I am the proc" << this->worldComm().localRank()<<" I receive to proc " << proc
                     <<" with tag "<< cpt
                     << " idRecv " << idRecv
                     << " it_ghost->G() " << mesh->element( idRecv ).G()
                     << std::endl;
#endif
            this->worldComm().localComm().send( proc, cpt, mesh->element( idRecv ).G() );
        }
    }

    typedef typename matrix_node<double>::type matrix_node_type;
    std::vector< std::vector<matrix_node_type> > getFinalInfoCheck( this->worldComm().localSize() );

    for ( int proc=0; proc<this->worldComm().localSize(); ++proc )
    {
        getFinalInfoCheck[proc].resize( nbMsgToSendCheck[proc] );

        for ( int cpt=0; cpt<nbMsgToSendCheck[proc]; ++cpt )
        {
            this->worldComm().localComm().recv( proc, cpt, getFinalInfoCheck[proc][cpt] );
#if 0
            std::cout<< "I am the proc " << this->worldComm().localRank()<<" I receive to proc " << proc
                     <<" with tag "<< cpt
                     << " points G " << getFinalInfoCheck[proc][cpt]
                     << std::endl;
#endif
        }
    }

    // check final data
    it_ghost=mesh->beginGhostElement();
    en_ghost=mesh->endGhostElement();

    for ( int cpt=0; it_ghost!=en_ghost; ++it_ghost,++cpt )
    {
        int/*auto*/ IdProcessOfGhost = it_ghost->processId();
        auto indic = mapMsgCheck[IdProcessOfGhost][cpt];
#if 0
        std::cout<< "FINAL CHECK! I am the proc" << this->worldComm().localRank()
                 << " G() initial " << it_ghost->G()
                 << " G() returned" << getFinalInfoCheck[IdProcessOfGhost/*idProc*/][indic]
                 << std::endl;
#endif
        // check test
        auto eltG=it_ghost->G();
        auto elt2G = getFinalInfoCheck[IdProcessOfGhost/*idProc*/][indic];

        for ( int i =0 ; i<eltG.size2(); i++ )
        {
            for ( int j =0 ; j<eltG.size1(); j++ )
                if ( eltG( i,j )!=elt2G( i,j ) ) std::cout << "\n BAD!!!!!!!!!!!!!!!!!!!!!!!!!!!!! " << std::endl;
        }
    }

#endif

}

} // Feel



#endif /* __ImporterGmsh_H */
